import ctypes
from edge import eventlog
import os
from PIL import Image
from glob import glob

MAX_WIDTH = 65535
MAX_HEIGHT = 65535


_LOG_LEVEL_NONE = 0
_LOG_LEVEL_ERROR = 1
_LOG_LEVEL_WARN = 2
_LOG_LEVEL_INFO = 3
_LOG_LEVEL_DEBUG = 4
_LOG_LEVEL_TRACE = 5

_LOG_LEVEL_NAME_TO_CODE_MAP = {
    'none': _LOG_LEVEL_NONE,  # ?
    'error': _LOG_LEVEL_ERROR,
    'warn': _LOG_LEVEL_WARN,
    'info': _LOG_LEVEL_INFO,
    'debug': _LOG_LEVEL_DEBUG,
    'trace': _LOG_LEVEL_TRACE
}

_FORMAT_NONE = 0
_FORMAT_MONO8 = 1
_FORMAT_YUV420_I420 = 5
_FORMAT_NV12 = 6
_FORMAT_YUV411 = 10
_FORMAT_UYVY = 15
_FORMAT_YUY2 = 16
_FORMAT_RGB = 20
_FORMAT_RGBA = 21
_FORMAT_RGBX = 22
_FORMAT_BAYER_RGGB = 30
_FORMAT_BAYER_GRBG = 31
_FORMAT_BAYER_BGGR = 32
_FORMAT_BAYER_GBRG = 33

_FORMAT_NAME_TO_CODE_MAP = {
    'MONO8T': _FORMAT_MONO8,
    'I420PLN': _FORMAT_YUV420_I420,
    'NV12PLN': _FORMAT_NV12,
    'YUV411T': _FORMAT_YUV411,
    'UYVY': _FORMAT_UYVY,
    'YUY2': _FORMAT_YUY2,
    'RGB': _FORMAT_RGB,
    'RGBA': _FORMAT_RGBA,
    'RGBX': _FORMAT_RGBX,
    'BayerRG8T': _FORMAT_BAYER_RGGB,
    'BayerGR8T': _FORMAT_BAYER_GRBG,
    'BayerBG8T': _FORMAT_BAYER_BGGR,
    'BayerGB8T': _FORMAT_BAYER_GBRG
}

_lib = ctypes.CDLL(glob('%s/librawconv*.so' % os.path.dirname(__file__))[0])


# int
# RawConvInit(int aLogLevel, int aSourceFormat, unsigned int aSourceWidth, unsigned int aSourceHeight,
#    int aTargetFormat, unsigned int aTargetWidth, unsigned int aTargetHeight)
_conv_init = _lib.RawConvInit
_conv_init.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_uint, ctypes.c_uint,
                       ctypes.c_int, ctypes.c_uint, ctypes.c_uint)
_conv_init.restype = ctypes.c_int


# int
# RawConvProcessFile(const char *aSourcePath, char *aTargetData, int *aTargetSize,
#    unsigned int *aTargetWidth, unsigned int *aTargetHeight);
_conv_process_file = _lib.RawConvProcessFile
_conv_process_file.argtypes = [ctypes.POINTER(ctypes.c_char), ctypes.POINTER(ctypes.c_char),
                               ctypes.POINTER(ctypes.c_int),
                               ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(ctypes.c_uint)]
_conv_process_file.restype = ctypes.c_int


# void RawConvFini()
_conv_fini = _lib.RawConvFini
_conv_fini.argtypes = []


class RawUtil:
    """
    :type _target_buffer: str
    """

    _NULL_SOURCE_FWH = ('', MAX_WIDTH, MAX_HEIGHT)
    _TARGET_FORMAT = 'RGBA'

    def __init__(self, target_max_wh, log_level):
        """
        :type target_max_wh: (int, int)
        :type log_level: int
        """
        self._source_fwh = RawUtil._NULL_SOURCE_FWH
        self._target_max_wh = target_max_wh

        self._target_buffer = None
        self._log_level = log_level

    @staticmethod
    def create(max_target_width=MAX_WIDTH, max_target_height=MAX_HEIGHT, log_level='info'):
        """
        :type max_target_width: int
        :type max_target_height: int
        :type log_level: str
        :rtype: RawUtil
        """
        target_max_wh = (max_target_width, max_target_height)
        eventlog.log_d('target-max-wh: %sx%s' % target_max_wh)

        log_level_code = _LOG_LEVEL_NAME_TO_CODE_MAP.get(log_level, _LOG_LEVEL_INFO)
        return RawUtil(target_max_wh, log_level_code)

    # noinspection PyMethodMayBeStatic
    def destroy(self):
        eventlog.log_d('try')
        if self._source_fwh != RawUtil._NULL_SOURCE_FWH:
            _conv_fini()
            eventlog.log_d('done: RawConvFini')

        eventlog.log_d('done')

    def handle_image3(self, source_url, source_format, source_width, source_height, target_buffer):
        """
        :type source_url: str
        :type source_format: str
        :type source_width: int
        :type source_height: int
        :type target_buffer: str
        :rtype: (str, int, int) | None
        """
        # check source-image format changed
        source_fwh = (source_format, source_width, source_height)
        if self._source_fwh != source_fwh:
            eventlog.log_d('%s -> %s' % (self._source_fwh, source_fwh))

            # resize to the bigger images is not supported
            target_max_w, target_max_h = self._target_max_wh
            target_max_w = source_width if source_width < target_max_w else target_max_w
            target_max_h = source_height if source_height < target_max_h else target_max_h

            ret_code = _conv_init(
                self._log_level, _FORMAT_NAME_TO_CODE_MAP.get(source_format), source_width, source_height,
                _FORMAT_NAME_TO_CODE_MAP.get(RawUtil._TARGET_FORMAT), target_max_w, target_max_h)

            if ret_code != 0:
                self._source_fwh = RawUtil._NULL_SOURCE_FWH
                eventlog.log_d('fail: RawConvInit (%s)' % ret_code)
                return None

            self._source_fwh = source_fwh

        # convert image
        target_width = ctypes.c_uint()
        target_height = ctypes.c_uint()
        target_size = ctypes.c_int(ctypes.sizeof(target_buffer))

        ret_code = _conv_process_file(
            source_url, ctypes.cast(target_buffer, ctypes.POINTER(ctypes.c_char)),
            target_size, target_width, target_height)

        if ret_code != 0:
            eventlog.log_d('fail: RawConvProcessFile (%s)' % ret_code)
            return None

        return target_buffer, target_width.value, target_height.value

    def handle_image2(self, source_url, source_format, source_width, source_height):
        """
        :type source_url: str
        :type source_format: str
        :type source_width: int
        :type source_height: int
        :rtype (str, int, int) | None
        """
        # check source-image format changed
        source_fwh = source_format, source_width, source_height
        if self._source_fwh != source_fwh:
            self._target_buffer = ctypes.create_string_buffer(4 * source_width * source_height)

        return self.handle_image3(source_url, source_format, source_width, source_height, self._target_buffer)

    def handle_image(self, source_url, source_format, source_width, source_height):
        """
        :type source_url: str
        :type source_format: str
        :type source_width: int
        :type source_height: int
        :rtype Image | None
        """
        target_bwh = self.handle_image2(source_url, source_format, source_width, source_height)
        if target_bwh is None:
            return None

        # attach buffer to pil-image
        target_buffer, target_width, target_height = target_bwh
        pil_image = Image.frombuffer(
            RawUtil._TARGET_FORMAT, (target_width, target_height),
            target_buffer,
            'raw', RawUtil._TARGET_FORMAT, 0, 1)

        pil_image.readonly = 0  # allow to write changes to attached image buffer
        return pil_image
